# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from ._yadisk_object import *
from ._error_object import *
from ._link_object import *
from ._disk import *
from ._resources import *
from ._operations import *
from ._auth import *

__all__ = [  # noqa: RUF022
    # _yadisk_object
    "YaDiskObject",

    # _error_object
    "ErrorObject",

    # _link_object
    "LinkObject",

    # _disk
    "DiskInfoObject",
    "SystemFoldersObject",
    "UserObject",
    "UserPublicInfoObject",

    # _resources
    "CommentIDsObject",
    "EXIFObject",
    "FilesResourceListObject",
    "SyncFilesResourceListObject",
    "AsyncFilesResourceListObject",
    "LastUploadedResourceListObject",
    "SyncLastUploadedResourceListObject",
    "AsyncLastUploadedResourceListObject",
    "PublicResourcesListObject",
    "SyncPublicResourcesListObject",
    "AsyncPublicResourcesListObject",
    "ResourceListObject",
    "SyncResourceListObject",
    "AsyncResourceListObject",
    "ResourceObject",
    "SyncResourceObject",
    "AsyncResourceObject",
    "ResourceUploadLinkObject",
    "ShareInfoObject",
    "PublicResourceObject",
    "SyncPublicResourceObject",
    "AsyncPublicResourceObject",
    "PublicResourceListObject",
    "SyncPublicResourceListObject",
    "AsyncPublicResourceListObject",
    "TrashResourceObject",
    "SyncTrashResourceObject",
    "AsyncTrashResourceObject",
    "TrashResourceListObject",
    "SyncTrashResourceListObject",
    "AsyncTrashResourceListObject",
    "ResourceLinkObject",
    "SyncResourceLinkObject",
    "AsyncResourceLinkObject",
    "PublicResourceLinkObject",
    "SyncPublicResourceLinkObject",
    "AsyncPublicResourceLinkObject",
    "ResourceDownloadLinkObject",

    # _operations
    "OperationStatusObject",
    "OperationLinkObject",
    "SyncOperationLinkObject",
    "AsyncOperationLinkObject",

    # _auth
    "TokenObject",
    "TokenRevokeStatusObject",
    "DeviceCodeObject"
]
